// -*- C++ -*-
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#include <cutil.h>

#define MAX_NUM_CLUSTERS (8)

void gpuKMeans (float *hx, float *hy, int point_count, int *hlabels, int cluster_count);

int checkNearest(float x, float y, float *avgx, float *avgy, int cluster_count) 
{
    int i;
    float nearest_distance = DBL_MAX;
    float tmp;
    int ret = 0;

    for (i = 0; i < cluster_count; i++) {
        tmp = sqrt(pow(x - avgx[i], 2)
                   + pow(y - avgy[i], 2));
        if (tmp < nearest_distance) {
            nearest_distance = tmp;
            ret = i;
        }
    }

    return ret;
}

void cpuKMeans(float *x, float *y, int point_count, int *labels, int cluster_count) 
{    
    int i;
    int loop_count = 0;
    int change;
    float *avgx, *avgy;
    float *sumx, *sumy;
    int *cnt;
        
    printf("[KMeans] start\n");

    avgx = (float *)malloc(cluster_count * sizeof(float));
    avgy = (float *)malloc(cluster_count * sizeof(float));
    sumx = (float *)malloc(cluster_count * sizeof(float));
    sumy = (float *)malloc(cluster_count * sizeof(float));
    cnt = (int *)malloc(cluster_count * sizeof(int));

    // [1]: average point of each cluster
    do {
        //        printf("[%3d LOOP]\n", loop_count);
        change = 0;
        for (i = 0; i <cluster_count; i++) {
            sumx[i] = 0.0;
            sumy[i] = 0.0;
            cnt[i] = 0;
        }
        for (i = 0; i < point_count; i++) {
            sumx[labels[i]] += x[i];
            sumy[labels[i]] += y[i];
            cnt[labels[i]]++;
        }
        
        for (i = 0; i < cluster_count; i++) {
            if (cnt[i] != 0) {
                avgx[i] = sumx[i] / cnt[i];
                avgy[i] = sumy[i] / cnt[i];
            }
        }


        // [2]: the nearest average point
        for (i = 0; i < point_count; i++) {
            int tmp;
            tmp = checkNearest(x[i], y[i], avgx, avgy, cluster_count);
            if (tmp != labels[i])
                change++;
            labels[i] = tmp;
        }
        loop_count++;
    } while (0 < change);
    // [3]: [1], [2] continue until no change in [2]

    printf("[KMeans] finish(total %d loops)\n", loop_count);

    free(sumx);
    free(sumy);
    free(cnt);
    free(avgx);
    free(avgy);
    return ;
}

int main(int argc, char *argv[]) {
    int i;
    int point_count;
    float tempx, tempy;
    FILE *infp, *outfp;
    int ret;
    float *x, *y;
    int *labels;
    unsigned int timer;
    float time;

    
    if (argc != 4) {
        fprintf(stderr, "ERROR: check Makefile\n");
        return -1;
    }

    infp = fopen(argv[1], "r");
    if (infp == NULL) {
        fprintf(stderr, "ERROR: failed to open %s\n", argv[1]);
        return -1;
    }

    outfp = fopen(argv[2], "w");
    if (outfp == NULL) {
        fprintf(stderr, "ERROR: failed to open %s\n", argv[2]);
        return -1;
    }

    // read input file to "points"
    ret = fscanf(infp, "%d", &point_count);
    if (ret < 0) {
        fprintf(stderr, "ERROR: failed to scan point_count\n");
        exit(1);
    }
    printf("The num of points: %d\n", point_count);
    x = (float *)malloc(point_count * sizeof(float));
    y = (float *)malloc(point_count * sizeof(float));
    labels = (int *)malloc(point_count * sizeof(int));
    for (i = 0; ; i++) {
        ret = fscanf(infp, "%f %f", &tempx, &tempy);
        if (ret == EOF)
            break;
	x[i] = tempx;
        y[i] = tempy;
    }
    if (i != point_count) {
        fprintf(stderr, "ERROR: wrong point_count in %s\n", argv[1]);
        exit(1);
    }
    for (i = 0; i < point_count; i++)
        labels[i] = i % MAX_NUM_CLUSTERS;

    // K-means processing
    // Your task is to optimize this function.
    cutCreateTimer(&timer);
    cutResetTimer(timer);
    cutStartTimer(timer);
    if (strcmp(argv[3], "-cpu") == 0)
      cpuKMeans(x, y, point_count, labels, MAX_NUM_CLUSTERS);
    else if (strcmp(argv[3], "-gpu") == 0)
      gpuKMeans(x, y, point_count, labels, MAX_NUM_CLUSTERS);
    cutStopTimer(timer);
    time = cutGetTimerValue(timer)*1.0e-03;
    
    // This processing time is your score.
    printf("[KMeans time]%11.4lf ms\n", time);

    // writing result to a file
    fprintf(outfp, "%d\n", MAX_NUM_CLUSTERS);
    for (i = 0; i < point_count; i++)
        fprintf(outfp, "%7.2lf %7.2lf %2d\n", x[i], y[i], labels[i]);

    free(x);
    free(labels);

    fclose(infp);
    fclose(outfp);

    return 0;
}

